# This sample tests the ability to detect errant assert calls
# that are always true - the "reportAssertAlwaysTrue" option.

from typing import Any, Tuple


# This should generate a warning.
assert (1 != 2, "Error message")


def foo(a: Tuple[int, ...]):
    assert a


b = ()
assert b


c = (2, 3)

# This should generate a warning.
assert c



import sys
from pathlib import Path
import shutil
import pyright_to_gitlab_ci.main as dut

TESTS = Path(__file__).resolve().parent
REPORTS = TESTS / "reports"
# pyright is expected to produce abs path, so put it all in /tmp
TMP = Path("/tmp")


def call_pyright_to_gitlab_ci(*argv):
    prev_sys_argv = sys.argv
    try:
        sys.argv = ["pyright-to-gitlab-ci"] + list(argv)
        dut.main()
    finally:
        sys.argv = prev_sys_argv


class TestReportsConversion:
    # automatically populated with the reports
    # in the reports/ directory.
    # each of these being a testcase with an
    # input and an output
    pass


def check_report_produced_correctly(
    f_source, f_input, f_expected_out, f_actual_out, base_path
):
    def _check(self):
        tmp_source = TMP / f_source.name
        shutil.copy(f_source, tmp_source)
        call_pyright_to_gitlab_ci(
            "--src",
            str(f_input),
            "--output",
            str(f_actual_out),
            "--base_path",
            str(base_path),
        )
        with open(f_expected_out) as f:
            expected_out = f.read()
        with open(f_actual_out) as f:
            actual_out = f.read()
        assert expected_out == actual_out

    return _check


for f_source in REPORTS.glob("*.py"):
    f_input = f_source.with_suffix(".pyright.json")
    f_expected_out = f_source.with_suffix(".expected.json")
    f_actual_out = TMP / f_source.with_suffix(".actual.json").name
    base_path = TMP
    if f_source.is_file() and f_expected_out.is_file():
        setattr(
            TestReportsConversion,
            f"test_reports_{f_input.with_suffix('').name}",
            check_report_produced_correctly(
                f_source, f_input, f_expected_out, f_actual_out, base_path
            ),
        )



import os
import sys
from pathlib import Path
from argparse import ArgumentParser
import json
from hashlib import md5
import typing
from typing import Any
from dataclasses import dataclass, asdict
from enum import Enum
import functools
import warnings


@dataclass(eq=True, frozen=True)
class CodeClimatePosition:
    line: int
    column: int


@dataclass(eq=True, frozen=True)
class CodeClimateRange:
    begin: CodeClimatePosition
    end: CodeClimatePosition


@dataclass
class PyrightPosition:
    line: int
    character: int


@dataclass
class PyrightRange:
    start: PyrightPosition
    end: PyrightPosition

    def __init__(self, start: dict[str, int], end: dict[str, int]):
        self.start = PyrightPosition(**start)
        self.end = PyrightPosition(**end)

    def to_positions(self) -> CodeClimateRange:
        return CodeClimateRange(
            begin=CodeClimatePosition(
                line=self.start.line, column=self.start.character
            ),
            end=CodeClimatePosition(line=self.end.line, column=self.end.character),
        )


@dataclass
class PyrightUri:
    _key: str
    _filePath: str


class PyrightSeverityLevel(str, Enum):
    error = "error"
    warning = "warning"
    information = "information"
    unusedcode = "unusedcode"
    unreachablecode = "unreachablecode"
    deprecated = "deprecated"


class GitlabSeverityLevel(str, Enum):
    blocker = "blocker"
    major = "major"
    minor = "minor"
    info = "info"


@dataclass(eq=True, frozen=True)
class GitlabReportLocation:
    path: str
    positions: CodeClimateRange

    @functools.cache
    def get_lines(self, base_path) -> str:
        with open(base_path / self.path, "rt") as f:
            ls = f.readlines()
        return "".join(ls[self.positions.begin.line : self.positions.end.line + 1])


@dataclass
class PyrightDiagnostic:
    severity: PyrightSeverityLevel
    message: str
    file: str | None = None
    uri: PyrightUri | None = None
    range: PyrightRange | None = None
    rule: str | None = None

    def extract_path(self) -> Path:
        if self.uri is not None:
            return Path(self.uri._filePath)
        elif self.file is not None:
            return Path(self.file)
        else:
            raise ValueError("BUG! no URI, no file?")

    def __init__(
        self,
        file: str | None = None,
        uri: dict[str, str] | None = None,
        range: dict[str, dict[str, int]] | None = None,
        **kw,
    ):
        assert file is not None or uri is not None

        self.file = file
        if uri is not None:
            self.uri = PyrightUri(**uri)
        else:
            self.uri = None

        if range is not None:
            self.range = PyrightRange(**range)
        else:
            self.range = None

        for n, v in kw.items():
            setattr(self, n, v)

    def maybe_gitlab_report_location(
        self, abs_base_path
    ) -> GitlabReportLocation | None:
        if self.range is not None:
            return GitlabReportLocation(
                path=str(self.extract_path().resolve().relative_to(abs_base_path)),
                positions=self.range.to_positions(),
            )


@dataclass
class PyrightReport:
    version: str
    time: str
    generalDiagnostics: list[PyrightDiagnostic]
    summary: dict[str, int]

    @classmethod
    def from_json_file(cls, f: typing.TextIO) -> "PyrightReport":
        return cls(**json.load(f))

    def __init__(self, generalDiagnostics: list[dict[str, Any]], **kw):
        self.generalDiagnostics = [PyrightDiagnostic(**o) for o in generalDiagnostics]
        for n, v in kw.items():
            setattr(self, n, v)


class ReportFingerprint(str):
    pass


@dataclass
class GitlabReport:
    description: str
    fingerprint: ReportFingerprint
    severity: GitlabSeverityLevel
    location: GitlabReportLocation


def fingerprint(
    base_path: Path, description: str, location: GitlabReportLocation
) -> ReportFingerprint:
    return ReportFingerprint(
        md5(
            b"\0".join((description.encode(), location.get_lines(base_path).encode())),
            usedforsecurity=False,
        ).hexdigest()
    )


@dataclass
class GitlabIncompleteReport:
    """
    Report info without the location and fingerprinting
    """

    description: str
    severity: GitlabSeverityLevel

    def complete(self, base_path: Path, location: GitlabReportLocation) -> GitlabReport:
        return GitlabReport(
            description=self.description,
            severity=self.severity,
            location=location,
            fingerprint=fingerprint(base_path, self.description, location),
        )


def load_json(fname: str) -> PyrightReport:
    with open(fname, "rt") as f:
        return PyrightReport.from_json_file(f)


def convert_diagnostic_category_to_gitlab_severity(
    category: PyrightSeverityLevel,
) -> GitlabSeverityLevel:
    match category:
        case PyrightSeverityLevel.error:
            return GitlabSeverityLevel.blocker
        case PyrightSeverityLevel.warning:
            return GitlabSeverityLevel.major
        case PyrightSeverityLevel.unreachablecode:
            return GitlabSeverityLevel.major
        case PyrightSeverityLevel.deprecated:
            return GitlabSeverityLevel.minor
        case _:
            return GitlabSeverityLevel.info


def main():
    ap = ArgumentParser()
    ap.add_argument("--src", type=load_json)
    ap.add_argument("--output", type=Path)
    ap.add_argument("--base_path", default=Path("."), type=Path)
    args = ap.parse_args()

    report = (
        args.src if args.src is not None else PyrightReport.from_json_file(sys.stdin)
    )

    if args.output is not None:
        args.output.parent.mkdir(exist_ok=True, parents=True)

    fout = open(args.output, "wt") if args.output is not None else sys.stdout

    response: list[GitlabReport] = []

    abs_base_path = args.base_path.resolve()

    for diag in report.generalDiagnostics:
        loc = diag.maybe_gitlab_report_location(abs_base_path)
        if loc is not None:
            response.append(
                GitlabIncompleteReport(
                    description=diag.message,
                    severity=convert_diagnostic_category_to_gitlab_severity(
                        diag.severity
                    ),
                ).complete(
                    base_path=abs_base_path,
                    location=loc,
                )
            )
        else:
            warnings.warn("some pyright reports thrown away because they had no range")

    print(json.dumps(list(asdict(r) for r in response)), file=fout)


if __name__ == "__main__":  # pragma: no cover
    main()
